@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2020 Canonical Ltd
+ * Copyright (C) 2020-2022 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -24,19 +24,19 @@ import (
 	"crypto/elliptic"
 	"crypto/rand"
 	"encoding/json"
-	"errors"
 	"fmt"
-	"io/ioutil"
 	"os"
 	"path/filepath"
 
 	"github.com/snapcore/snapd/asserts"
 	"github.com/snapcore/snapd/bootloader"
 	"github.com/snapcore/snapd/dirs"
+	"github.com/snapcore/snapd/gadget/device"
 	"github.com/snapcore/snapd/kernel/fde"
 	"github.com/snapcore/snapd/logger"
 	"github.com/snapcore/snapd/osutil"
 	"github.com/snapcore/snapd/secboot"
+	"github.com/snapcore/snapd/secboot/keys"
 	"github.com/snapcore/snapd/seed"
 	"github.com/snapcore/snapd/snap"
 	"github.com/snapcore/snapd/strutil"
@@ -44,9 +44,12 @@ import (
 )
 
 var (
-	secbootSealKeys                 = secboot.SealKeys
-	secbootSealKeysWithFDESetupHook = secboot.SealKeysWithFDESetupHook
-	secbootResealKeys               = secboot.ResealKeys
+	secbootProvisionTPM              = secboot.ProvisionTPM
+	secbootSealKeys                  = secboot.SealKeys
+	secbootSealKeysWithFDESetupHook  = secboot.SealKeysWithFDESetupHook
+	secbootResealKeys                = secboot.ResealKeys
+	secbootPCRHandleOfSealedKey      = secboot.PCRHandleOfSealedKey
+	secbootReleasePCRResourceHandles = secboot.ReleasePCRResourceHandles
 
 	seedReadSystemEssential = seed.ReadSystemEssential
 )
@@ -55,7 +58,10 @@ var (
 // disk encryption implementations. The state must be locked when these
 // functions are called.
 var (
-	HasFDESetupHook = func() (bool, error) {
+	// HasFDESetupHook purpose is to detect if the target kernel has a
+	// fde-setup-hook. If kernelInfo is nil the current kernel is checked
+	// assuming it is representative` of the target one.
+	HasFDESetupHook = func(kernelInfo *snap.Info) (bool, error) {
 		return false, nil
 	}
 	RunFDESetupHook fde.RunSetupHookFunc = func(req *fde.SetupRequest) ([]byte, error) {
@@ -63,14 +69,6 @@ var (
 	}
 )
 
-type sealingMethod string
-
-const (
-	sealingMethodLegacyTPM    = sealingMethod("")
-	sealingMethodTPM          = sealingMethod("tpm")
-	sealingMethodFDESetupHook = sealingMethod("fde-setup-hook")
-)
-
 // MockSecbootResealKeys is only useful in testing. Note that this is a very low
 // level call and may need significant environment setup.
 func MockSecbootResealKeys(f func(params *secboot.ResealKeysParams) error) (restore func()) {
@@ -92,6 +90,18 @@ func MockResealKeyToModeenv(f func(rootdir string, modeenv *Modeenv, expectResea
 	}
 }
 
+// MockSealKeyToModeenvFlags is used for testing from other packages.
+type MockSealKeyToModeenvFlags = sealKeyToModeenvFlags
+
+// MockSealKeyToModeenv is used for testing from other packages.
+func MockSealKeyToModeenv(f func(key, saveKey keys.EncryptionKey, model *asserts.Model, modeenv *Modeenv, flags MockSealKeyToModeenvFlags) error) (restore func()) {
+	old := sealKeyToModeenv
+	sealKeyToModeenv = f
+	return func() {
+		sealKeyToModeenv = old
+	}
+}
+
 func bootChainsFileUnder(rootdir string) string {
 	return filepath.Join(dirs.SnapFDEDirUnder(rootdir), "boot-chains")
 }
@@ -100,15 +110,26 @@ func recoveryBootChainsFileUnder(rootdir string) string {
 	return filepath.Join(dirs.SnapFDEDirUnder(rootdir), "recovery-boot-chains")
 }
 
-// sealKeyToModeenv seals the supplied keys to the parameters specified
+type sealKeyToModeenvFlags struct {
+	// HasFDESetupHook is true if the kernel has a fde-setup hook to use
+	HasFDESetupHook bool
+	// FactoryReset indicates that the sealing is happening during factory
+	// reset.
+	FactoryReset bool
+	// SnapsDir is set to provide a non-default directory to find
+	// run mode snaps in.
+	SnapsDir string
+}
+
+// sealKeyToModeenvImpl seals the supplied keys to the parameters specified
 // in modeenv.
 // It assumes to be invoked in install mode.
-func sealKeyToModeenv(key, saveKey secboot.EncryptionKey, model *asserts.Model, modeenv *Modeenv) error {
+func sealKeyToModeenvImpl(key, saveKey keys.EncryptionKey, model *asserts.Model, modeenv *Modeenv, flags sealKeyToModeenvFlags) error {
 	// make sure relevant locations exist
 	for _, p := range []string{
 		InitramfsSeedEncryptionKeyDir,
 		InitramfsBootEncryptionKeyDir,
-		InstallHostFDEDataDir,
+		InstallHostFDEDataDir(model),
 		InstallHostFDESaveDir,
 	} {
 		// XXX: should that be 0700 ?
@@ -117,50 +138,54 @@ func sealKeyToModeenv(key, saveKey secboot.EncryptionKey, model *asserts.Model,
 		}
 	}
 
-	hasHook, err := HasFDESetupHook()
-	if err != nil {
-		return fmt.Errorf("cannot check for fde-setup hook %v", err)
-	}
-	if hasHook {
-		return sealKeyToModeenvUsingFDESetupHook(key, saveKey, modeenv)
+	if flags.HasFDESetupHook {
+		return sealKeyToModeenvUsingFDESetupHook(key, saveKey, model, modeenv, flags)
 	}
 
-	return sealKeyToModeenvUsingSecboot(key, saveKey, modeenv)
+	return sealKeyToModeenvUsingSecboot(key, saveKey, model, modeenv, flags)
 }
 
-func runKeySealRequests(key secboot.EncryptionKey) []secboot.SealKeyRequest {
+func runKeySealRequests(key keys.EncryptionKey) []secboot.SealKeyRequest {
 	return []secboot.SealKeyRequest{
 		{
 			Key:     key,
 			KeyName: "ubuntu-data",
-			KeyFile: filepath.Join(InitramfsBootEncryptionKeyDir, "ubuntu-data.sealed-key"),
+			KeyFile: device.DataSealedKeyUnder(InitramfsBootEncryptionKeyDir),
 		},
 	}
 }
 
-func fallbackKeySealRequests(key, saveKey secboot.EncryptionKey) []secboot.SealKeyRequest {
+func fallbackKeySealRequests(key, saveKey keys.EncryptionKey, factoryReset bool) []secboot.SealKeyRequest {
+	saveFallbackKey := device.FallbackSaveSealedKeyUnder(InitramfsSeedEncryptionKeyDir)
+
+	if factoryReset {
+		// factory reset uses alternative sealed key location, such that
+		// until we boot into the run mode, both sealed keys are present
+		// on disk
+		saveFallbackKey = device.FactoryResetFallbackSaveSealedKeyUnder(InitramfsSeedEncryptionKeyDir)
+	}
 	return []secboot.SealKeyRequest{
 		{
 			Key:     key,
 			KeyName: "ubuntu-data",
-			KeyFile: filepath.Join(InitramfsSeedEncryptionKeyDir, "ubuntu-data.recovery.sealed-key"),
+			KeyFile: device.FallbackDataSealedKeyUnder(InitramfsSeedEncryptionKeyDir),
 		},
 		{
 			Key:     saveKey,
 			KeyName: "ubuntu-save",
-			KeyFile: filepath.Join(InitramfsSeedEncryptionKeyDir, "ubuntu-save.recovery.sealed-key"),
+			KeyFile: saveFallbackKey,
 		},
 	}
 }
 
-func sealKeyToModeenvUsingFDESetupHook(key, saveKey secboot.EncryptionKey, modeenv *Modeenv) error {
+func sealKeyToModeenvUsingFDESetupHook(key, saveKey keys.EncryptionKey, model *asserts.Model, modeenv *Modeenv, flags sealKeyToModeenvFlags) error {
 	// XXX: Move the auxKey creation to a more generic place, see
 	// PR#10123 for a possible way of doing this. However given
 	// that the equivalent key for the TPM case is also created in
 	// sealKeyToModeenvUsingTPM more symetric to create the auxKey
 	// here and when we also move TPM to use the auxKey to move
 	// the creation of it.
-	auxKey, err := secboot.NewAuxKey()
+	auxKey, err := keys.NewAuxKey()
 	if err != nil {
 		return fmt.Errorf("cannot create aux key: %v", err)
 	}
@@ -169,19 +194,20 @@ func sealKeyToModeenvUsingFDESetupHook(key, saveKey secboot.EncryptionKey, modee
 		AuxKey:     auxKey,
 		AuxKeyFile: filepath.Join(InstallHostFDESaveDir, "aux-key"),
 	}
-	skrs := append(runKeySealRequests(key), fallbackKeySealRequests(key, saveKey)...)
+	factoryReset := flags.FactoryReset
+	skrs := append(runKeySealRequests(key), fallbackKeySealRequests(key, saveKey, factoryReset)...)
 	if err := secbootSealKeysWithFDESetupHook(RunFDESetupHook, skrs, &params); err != nil {
 		return err
 	}
 
-	if err := stampSealedKeys(InstallHostWritableDir, "fde-setup-hook"); err != nil {
+	if err := device.StampSealedKeys(InstallHostWritableDir(model), "fde-setup-hook"); err != nil {
 		return err
 	}
 
 	return nil
 }
 
-func sealKeyToModeenvUsingSecboot(key, saveKey secboot.EncryptionKey, modeenv *Modeenv) error {
+func sealKeyToModeenvUsingSecboot(key, saveKey keys.EncryptionKey, model *asserts.Model, modeenv *Modeenv, flags sealKeyToModeenvFlags) error {
 	// build the recovery mode boot chain
 	rbl, err := bootloader.Find(InitramfsUbuntuSeedDir, &bootloader.Options{
 		Role: bootloader.RoleRecovery,
@@ -196,10 +222,17 @@ func sealKeyToModeenvUsingSecboot(key, saveKey secboot.EncryptionKey, modeenv *M
 	}
 
 	includeTryModel := false
-	recoveryBootChains, err := recoveryBootChainsForSystems([]string{modeenv.RecoverySystem}, tbl, modeenv, includeTryModel)
+	systems := []string{modeenv.RecoverySystem}
+	modes := map[string][]string{
+		// the system we are installing from is considered current and
+		// tested, hence allow both recover and factory reset modes
+		modeenv.RecoverySystem: {ModeRecover, ModeFactoryReset},
+	}
+	recoveryBootChains, err := recoveryBootChainsForSystems(systems, modes, tbl, modeenv, includeTryModel)
 	if err != nil {
 		return fmt.Errorf("cannot compose recovery boot chains: %v", err)
 	}
+	logger.Debugf("recovery bootchain:\n%+v", recoveryBootChains)
 
 	// build the run mode boot chains
 	bl, err := bootloader.Find(InitramfsUbuntuBootDir, &bootloader.Options{
@@ -212,10 +245,11 @@ func sealKeyToModeenvUsingSecboot(key, saveKey secboot.EncryptionKey, modeenv *M
 
 	// kernel command lines are filled during install
 	cmdlines := modeenv.CurrentKernelCommandLines
-	runModeBootChains, err := runModeBootChains(rbl, bl, modeenv, cmdlines)
+	runModeBootChains, err := runModeBootChains(rbl, bl, modeenv, cmdlines, flags.SnapsDir)
 	if err != nil {
 		return fmt.Errorf("cannot compose run mode boot chains: %v", err)
 	}
+	logger.Debugf("run mode bootchain:\n%+v", runModeBootChains)
 
 	pbc := toPredictableBootChains(append(runModeBootChains, recoveryBootChains...))
 
@@ -233,24 +267,71 @@ func sealKeyToModeenvUsingSecboot(key, saveKey secboot.EncryptionKey, modeenv *M
 		return fmt.Errorf("cannot generate key for signing dynamic authorization policies: %v", err)
 	}
 
-	if err := sealRunObjectKeys(key, pbc, authKey, roleToBlName); err != nil {
+	runObjectKeyPCRHandle := uint32(secboot.RunObjectPCRPolicyCounterHandle)
+	fallbackObjectKeyPCRHandle := uint32(secboot.FallbackObjectPCRPolicyCounterHandle)
+	if flags.FactoryReset {
+		// during factory reset we may need to rotate the PCR handles,
+		// seal the new keys using a new set of handles such that the
+		// old sealed ubuntu-save key is still usable, for this we
+		// switch between two sets of handles in a round robin fashion,
+		// first looking at the PCR handle used by the current fallback
+		// key and then using the other set when sealing the new keys;
+		// the currently used handles will be released during the first
+		// boot of a new run system
+		usesAlt, err := usesAltPCRHandles()
+		if err != nil {
+			return err
+		}
+		if !usesAlt {
+			logger.Noticef("using alternative PCR handles")
+			runObjectKeyPCRHandle = secboot.AltRunObjectPCRPolicyCounterHandle
+			fallbackObjectKeyPCRHandle = secboot.AltFallbackObjectPCRPolicyCounterHandle
+		}
+	}
+
+	// we are preparing a new system, hence the TPM needs to be provisioned
+	lockoutAuthFile := device.TpmLockoutAuthUnder(InstallHostFDESaveDir)
+	tpmProvisionMode := secboot.TPMProvisionFull
+	if flags.FactoryReset {
+		tpmProvisionMode = secboot.TPMPartialReprovision
+	}
+	if err := secbootProvisionTPM(tpmProvisionMode, lockoutAuthFile); err != nil {
 		return err
 	}
 
-	if err := sealFallbackObjectKeys(key, saveKey, rpbc, authKey, roleToBlName); err != nil {
+	if flags.FactoryReset {
+		// it is possible that we are sealing the keys again, after a
+		// previously running factory reset was interrupted by a reboot,
+		// in which case the PCR handles of the new sealed keys might
+		// have already been used
+		if err := secbootReleasePCRResourceHandles(runObjectKeyPCRHandle, fallbackObjectKeyPCRHandle); err != nil {
+			return err
+		}
+	}
+
+	// TODO: refactor sealing functions to take a struct instead of so many
+	// parameters
+	err = sealRunObjectKeys(key, pbc, authKey, roleToBlName, runObjectKeyPCRHandle)
+	if err != nil {
 		return err
 	}
 
-	if err := stampSealedKeys(InstallHostWritableDir, sealingMethodTPM); err != nil {
+	err = sealFallbackObjectKeys(key, saveKey, rpbc, authKey, roleToBlName, flags.FactoryReset,
+		fallbackObjectKeyPCRHandle)
+	if err != nil {
 		return err
 	}
 
-	installBootChainsPath := bootChainsFileUnder(InstallHostWritableDir)
+	if err := device.StampSealedKeys(InstallHostWritableDir(model), device.SealingMethodTPM); err != nil {
+		return err
+	}
+
+	installBootChainsPath := bootChainsFileUnder(InstallHostWritableDir(model))
 	if err := writeBootChains(pbc, installBootChainsPath, 0); err != nil {
 		return err
 	}
 
-	installRecoveryBootChainsPath := recoveryBootChainsFileUnder(InstallHostWritableDir)
+	installRecoveryBootChainsPath := recoveryBootChainsFileUnder(InstallHostWritableDir(model))
 	if err := writeBootChains(rpbc, installRecoveryBootChainsPath, 0); err != nil {
 		return err
 	}
@@ -258,7 +339,18 @@ func sealKeyToModeenvUsingSecboot(key, saveKey secboot.EncryptionKey, modeenv *M
 	return nil
 }
 
-func sealRunObjectKeys(key secboot.EncryptionKey, pbc predictableBootChains, authKey *ecdsa.PrivateKey, roleToBlName map[bootloader.Role]string) error {
+func usesAltPCRHandles() (bool, error) {
+	saveFallbackKey := device.FallbackSaveSealedKeyUnder(InitramfsSeedEncryptionKeyDir)
+	// inspect the PCR handle of the ubuntu-save fallback key
+	handle, err := secbootPCRHandleOfSealedKey(saveFallbackKey)
+	if err != nil {
+		return false, err
+	}
+	logger.Noticef("fallback sealed key %v PCR handle: %#x", saveFallbackKey, handle)
+	return handle == secboot.AltFallbackObjectPCRPolicyCounterHandle, nil
+}
+
+func sealRunObjectKeys(key keys.EncryptionKey, pbc predictableBootChains, authKey *ecdsa.PrivateKey, roleToBlName map[bootloader.Role]string, pcrHandle uint32) error {
 	modelParams, err := sealKeyModelParams(pbc, roleToBlName)
 	if err != nil {
 		return fmt.Errorf("cannot prepare for key sealing: %v", err)
@@ -268,10 +360,10 @@ func sealRunObjectKeys(key secboot.EncryptionKey, pbc predictableBootChains, aut
 		ModelParams:            modelParams,
 		TPMPolicyAuthKey:       authKey,
 		TPMPolicyAuthKeyFile:   filepath.Join(InstallHostFDESaveDir, "tpm-policy-auth-key"),
-		TPMLockoutAuthFile:     filepath.Join(InstallHostFDESaveDir, "tpm-lockout-auth"),
-		TPMProvision:           true,
-		PCRPolicyCounterHandle: secboot.RunObjectPCRPolicyCounterHandle,
+		PCRPolicyCounterHandle: pcrHandle,
 	}
+
+	logger.Debugf("sealing run key with PCR handle: %#x", sealKeyParams.PCRPolicyCounterHandle)
 	// The run object contains only the ubuntu-data key; the ubuntu-save key
 	// is then stored inside the encrypted data partition, so that the normal run
 	// path only unseals one object because unsealing is expensive.
@@ -284,7 +376,7 @@ func sealRunObjectKeys(key secboot.EncryptionKey, pbc predictableBootChains, aut
 	return nil
 }
 
-func sealFallbackObjectKeys(key, saveKey secboot.EncryptionKey, pbc predictableBootChains, authKey *ecdsa.PrivateKey, roleToBlName map[bootloader.Role]string) error {
+func sealFallbackObjectKeys(key, saveKey keys.EncryptionKey, pbc predictableBootChains, authKey *ecdsa.PrivateKey, roleToBlName map[bootloader.Role]string, factoryReset bool, pcrHandle uint32) error {
 	// also seal the keys to the recovery bootchains as a fallback
 	modelParams, err := sealKeyModelParams(pbc, roleToBlName)
 	if err != nil {
@@ -293,44 +385,20 @@ func sealFallbackObjectKeys(key, saveKey secboot.EncryptionKey, pbc predictableB
 	sealKeyParams := &secboot.SealKeysParams{
 		ModelParams:            modelParams,
 		TPMPolicyAuthKey:       authKey,
-		PCRPolicyCounterHandle: secboot.FallbackObjectPCRPolicyCounterHandle,
+		PCRPolicyCounterHandle: pcrHandle,
 	}
+	logger.Debugf("sealing fallback key with PCR handle: %#x", sealKeyParams.PCRPolicyCounterHandle)
 	// The fallback object contains the ubuntu-data and ubuntu-save keys. The
 	// key files are stored on ubuntu-seed, separate from ubuntu-data so they
 	// can be used if ubuntu-data and ubuntu-boot are corrupted or unavailable.
-	if err := secbootSealKeys(fallbackKeySealRequests(key, saveKey), sealKeyParams); err != nil {
+
+	if err := secbootSealKeys(fallbackKeySealRequests(key, saveKey, factoryReset), sealKeyParams); err != nil {
 		return fmt.Errorf("cannot seal the fallback encryption keys: %v", err)
 	}
 
 	return nil
 }
 
-func stampSealedKeys(rootdir string, content sealingMethod) error {
-	stamp := filepath.Join(dirs.SnapFDEDirUnder(rootdir), "sealed-keys")
-	if err := os.MkdirAll(filepath.Dir(stamp), 0755); err != nil {
-		return fmt.Errorf("cannot create device fde state directory: %v", err)
-	}
-
-	if err := osutil.AtomicWriteFile(stamp, []byte(content), 0644, 0); err != nil {
-		return fmt.Errorf("cannot create fde sealed keys stamp file: %v", err)
-	}
-	return nil
-}
-
-var errNoSealedKeys = errors.New("no sealed keys")
-
-// sealedKeysMethod return whether any keys were sealed at all
-func sealedKeysMethod(rootdir string) (sm sealingMethod, err error) {
-	// TODO:UC20: consider more than the marker for cases where we reseal
-	// outside of run mode
-	stamp := filepath.Join(dirs.SnapFDEDirUnder(rootdir), "sealed-keys")
-	content, err := ioutil.ReadFile(stamp)
-	if os.IsNotExist(err) {
-		return sm, errNoSealedKeys
-	}
-	return sealingMethod(content), err
-}
-
 var resealKeyToModeenv = resealKeyToModeenvImpl
 
 // resealKeyToModeenv reseals the existing encryption key to the
@@ -341,8 +409,8 @@ var resealKeyToModeenv = resealKeyToModeenvImpl
 // transient/in-memory information with the risk that successive
 // reseals during in-progress operations produce diverging outcomes.
 func resealKeyToModeenvImpl(rootdir string, modeenv *Modeenv, expectReseal bool) error {
-	method, err := sealedKeysMethod(rootdir)
-	if err == errNoSealedKeys {
+	method, err := device.SealedKeysMethod(rootdir)
+	if err == device.ErrNoSealedKeys {
 		// nothing to do
 		return nil
 	}
@@ -350,9 +418,9 @@ func resealKeyToModeenvImpl(rootdir string, modeenv *Modeenv, expectReseal bool)
 		return err
 	}
 	switch method {
-	case sealingMethodFDESetupHook:
+	case device.SealingMethodFDESetupHook:
 		return resealKeyToModeenvUsingFDESetupHook(rootdir, modeenv, expectReseal)
-	case sealingMethodTPM, sealingMethodLegacyTPM:
+	case device.SealingMethodTPM, device.SealingMethodLegacyTPM:
 		return resealKeyToModeenvSecboot(rootdir, modeenv, expectReseal)
 	default:
 		return fmt.Errorf("unknown key sealing method: %q", method)
@@ -396,13 +464,15 @@ func resealKeyToModeenvSecboot(rootdir string, modeenv *Modeenv, expectReseal bo
 		// TODO:UC20: later the exact kind of bootloaders we expect here might change
 		return fmt.Errorf("internal error: sealed keys but not a trusted assets bootloader")
 	}
+	// derive the allowed modes for each system mentioned in the modeenv
+	modes := modesForSystems(modeenv)
 
 	// the recovery boot chains for the run key are generated for all
 	// recovery systems, including those that are being tried; since this is
 	// a run key, the boot chains are generated for both models to
 	// accommodate the dynamics of a remodel
 	includeTryModel := true
-	recoveryBootChainsForRunKey, err := recoveryBootChainsForSystems(modeenv.CurrentRecoverySystems, tbl,
+	recoveryBootChainsForRunKey, err := recoveryBootChainsForSystems(modeenv.CurrentRecoverySystems, modes, tbl,
 		modeenv, includeTryModel)
 	if err != nil {
 		return fmt.Errorf("cannot compose recovery boot chains for run key: %v", err)
@@ -421,7 +491,7 @@ func resealKeyToModeenvSecboot(rootdir string, modeenv *Modeenv, expectReseal bo
 	// use the current model as the recovery keys are not expected to be
 	// used during a remodel
 	includeTryModel = false
-	recoveryBootChains, err := recoveryBootChainsForSystems(testedRecoverySystems, tbl, modeenv, includeTryModel)
+	recoveryBootChains, err := recoveryBootChainsForSystems(testedRecoverySystems, modes, tbl, modeenv, includeTryModel)
 	if err != nil {
 		return fmt.Errorf("cannot compose recovery boot chains: %v", err)
 	}
@@ -438,7 +508,7 @@ func resealKeyToModeenvSecboot(rootdir string, modeenv *Modeenv, expectReseal bo
 	if err != nil {
 		return err
 	}
-	runModeBootChains, err := runModeBootChains(rbl, bl, modeenv, cmdlines)
+	runModeBootChains, err := runModeBootChains(rbl, bl, modeenv, cmdlines, "")
 	if err != nil {
 		return fmt.Errorf("cannot compose run mode boot chains: %v", err)
 	}
@@ -510,9 +580,7 @@ func resealRunObjectKeys(pbc predictableBootChains, authKeyFile string, roleToBl
 	}
 
 	// list all the key files to reseal
-	keyFiles := []string{
-		filepath.Join(InitramfsBootEncryptionKeyDir, "ubuntu-data.sealed-key"),
-	}
+	keyFiles := []string{device.DataSealedKeyUnder(InitramfsBootEncryptionKeyDir)}
 
 	resealKeyParams := &secboot.ResealKeysParams{
 		ModelParams:          modelParams,
@@ -535,8 +603,8 @@ func resealFallbackObjectKeys(pbc predictableBootChains, authKeyFile string, rol
 
 	// list all the key files to reseal
 	keyFiles := []string{
-		filepath.Join(InitramfsSeedEncryptionKeyDir, "ubuntu-data.recovery.sealed-key"),
-		filepath.Join(InitramfsSeedEncryptionKeyDir, "ubuntu-save.recovery.sealed-key"),
+		device.FallbackDataSealedKeyUnder(InitramfsSeedEncryptionKeyDir),
+		device.FallbackSaveSealedKeyUnder(InitramfsSeedEncryptionKeyDir),
 	}
 
 	resealKeyParams := &secboot.ResealKeysParams{
@@ -551,10 +619,36 @@ func resealFallbackObjectKeys(pbc predictableBootChains, authKeyFile string, rol
 	return nil
 }
 
+// recoveryModesForSystems returns a map for recovery modes for recovery systems
+// mentioned in the modeenv. The returned map contains both tested and candidate
+// recovery systems
+func modesForSystems(modeenv *Modeenv) map[string][]string {
+	if len(modeenv.GoodRecoverySystems) == 0 && len(modeenv.CurrentRecoverySystems) == 0 {
+		return nil
+	}
+
+	systemToModes := map[string][]string{}
+
+	// first go through tested recovery systems
+	modesForTestedSystem := []string{ModeRecover, ModeFactoryReset}
+	// tried systems can only boot to recovery mode
+	modesForCandidateSystem := []string{ModeRecover}
+
+	// go through current recovery systems which can contain both tried
+	// systems and candidate ones
+	for _, sys := range modeenv.CurrentRecoverySystems {
+		systemToModes[sys] = modesForCandidateSystem
+	}
+	// go through recovery systems that were tested and update their modes
+	for _, sys := range modeenv.GoodRecoverySystems {
+		systemToModes[sys] = modesForTestedSystem
+	}
+	return systemToModes
+}
+
 // TODO:UC20: this needs to take more than one model to accommodate the remodel
 // scenario
-func recoveryBootChainsForSystems(systems []string, trbl bootloader.TrustedAssetsBootloader, modeenv *Modeenv, includeTryModel bool) (chains []bootChain, err error) {
-
+func recoveryBootChainsForSystems(systems []string, modesForSystems map[string][]string, trbl bootloader.TrustedAssetsBootloader, modeenv *Modeenv, includeTryModel bool) (chains []bootChain, err error) {
 	chainsForModel := func(model secboot.ModelForSealing) error {
 		modelID := modelUniqueID(model)
 		for _, system := range systems {
@@ -583,10 +677,18 @@ func recoveryBootChainsForSystems(systems []string, trbl bootloader.TrustedAsset
 				seedKernel, seedGadget = seedGadget, seedKernel
 			}
 
-			// get the command line
-			cmdline, err := composeCommandLine(currentEdition, ModeRecover, system, seedGadget.Path)
-			if err != nil {
-				return fmt.Errorf("cannot obtain recovery kernel command line: %v", err)
+			var cmdlines []string
+			modes, ok := modesForSystems[system]
+			if !ok {
+				return fmt.Errorf("internal error: no modes for system %q", system)
+			}
+			for _, mode := range modes {
+				// get the command line for this mode
+				cmdline, err := composeCommandLine(currentEdition, mode, system, seedGadget.Path)
+				if err != nil {
+					return fmt.Errorf("cannot obtain kernel command line for mode %q: %v", mode, err)
+				}
+				cmdlines = append(cmdlines, cmdline)
 			}
 
 			var kernelRev string
@@ -606,14 +708,16 @@ func recoveryBootChainsForSystems(systems []string, trbl bootloader.TrustedAsset
 			}
 
 			chains = append(chains, bootChain{
-				BrandID:        model.BrandID(),
-				Model:          model.Model(),
+				BrandID: model.BrandID(),
+				Model:   model.Model(),
+				// TODO: test this
+				Classic:        model.Classic(),
 				Grade:          model.Grade(),
 				ModelSignKeyID: model.SignKeyID(),
 				AssetChain:     assetChain,
 				Kernel:         seedKernel.SnapName(),
 				KernelRevision: kernelRev,
-				KernelCmdlines: []string{cmdline},
+				KernelCmdlines: cmdlines,
 				kernelBootFile: kbf,
 			})
 		}
@@ -633,7 +737,7 @@ func recoveryBootChainsForSystems(systems []string, trbl bootloader.TrustedAsset
 	return chains, nil
 }
 
-func runModeBootChains(rbl, bl bootloader.Bootloader, modeenv *Modeenv, cmdlines []string) ([]bootChain, error) {
+func runModeBootChains(rbl, bl bootloader.Bootloader, modeenv *Modeenv, cmdlines []string, runSnapsDir string) ([]bootChain, error) {
 	tbl, ok := rbl.(bootloader.TrustedAssetsBootloader)
 	if !ok {
 		return nil, fmt.Errorf("recovery bootloader doesn't support trusted assets")
@@ -646,7 +750,13 @@ func runModeBootChains(rbl, bl bootloader.Bootloader, modeenv *Modeenv, cmdlines
 			if err != nil {
 				return err
 			}
-			runModeBootChain, err := tbl.BootChain(bl, info.MountFile())
+			var kernelPath string
+			if runSnapsDir == "" {
+				kernelPath = info.MountFile()
+			} else {
+				kernelPath = filepath.Join(runSnapsDir, info.Filename())
+			}
+			runModeBootChain, err := tbl.BootChain(bl, kernelPath)
 			if err != nil {
 				return err
 			}
@@ -661,8 +771,10 @@ func runModeBootChains(rbl, bl bootloader.Bootloader, modeenv *Modeenv, cmdlines
 				kernelRev = info.SnapRevision().String()
 			}
 			chains = append(chains, bootChain{
-				BrandID:        model.BrandID(),
-				Model:          model.Model(),
+				BrandID: model.BrandID(),
+				Model:   model.Model(),
+				// TODO: test this
+				Classic:        model.Classic(),
 				Grade:          model.Grade(),
 				ModelSignKeyID: model.SignKeyID(),
 				AssetChain:     assetChain,
@@ -773,3 +885,61 @@ func isResealNeeded(pbc predictableBootChains, bootChainsFile string, expectRese
 	}
 	return true, c + 1, nil
 }
+
+func postFactoryResetCleanupSecboot() error {
+	// we are inspecting a key which was generated during factory reset, in
+	// the simplest case the sealed key generated previously used the main
+	// handles, while the current key uses alt handles, hence we need to
+	// release the main handles corresponding to the old key
+	handles := []uint32{secboot.RunObjectPCRPolicyCounterHandle, secboot.FallbackObjectPCRPolicyCounterHandle}
+	usesAlt, err := usesAltPCRHandles()
+	if err != nil {
+		return fmt.Errorf("cannot inspect fallback key: %v", err)
+	}
+	if !usesAlt {
+		// current fallback key using the main handles, which is
+		// possible of there were subsequent factory reset steps,
+		// release the alt handles associated with the old key
+		handles = []uint32{secboot.AltRunObjectPCRPolicyCounterHandle, secboot.AltFallbackObjectPCRPolicyCounterHandle}
+	}
+	return secbootReleasePCRResourceHandles(handles...)
+}
+
+func postFactoryResetCleanup() error {
+	hasHook, err := HasFDESetupHook(nil)
+	if err != nil {
+		return fmt.Errorf("cannot check for fde-setup hook %v", err)
+	}
+
+	saveFallbackKeyFactory := device.FactoryResetFallbackSaveSealedKeyUnder(InitramfsSeedEncryptionKeyDir)
+	saveFallbackKey := device.FallbackSaveSealedKeyUnder(InitramfsSeedEncryptionKeyDir)
+	if err := os.Rename(saveFallbackKeyFactory, saveFallbackKey); err != nil {
+		// it is possible that the key file was already renamed if we
+		// came back here after an unexpected reboot
+		if !os.IsNotExist(err) {
+			return fmt.Errorf("cannot rotate fallback key: %v", err)
+		}
+	}
+
+	if hasHook {
+		// TODO: do we need to invoke FDE hook?
+		return nil
+	}
+
+	if err := postFactoryResetCleanupSecboot(); err != nil {
+		return fmt.Errorf("cannot cleanup secboot state: %v", err)
+	}
+
+	return nil
+}
+
+// resealExpectedByModeenvChange returns true if resealing is expected
+// due to modeenv changes, false otherwise. Reseal might not be needed
+// if the only change in modeenv is the gadget (if the boot assets
+// change that is detected in resealKeyToModeenv() and reseal will
+// happen anyway)
+func resealExpectedByModeenvChange(m1, m2 *Modeenv) bool {
+	auxModeenv := *m2
+	auxModeenv.Gadget = m1.Gadget
+	return !auxModeenv.deepEqual(m1)
+}
