@@ -28,19 +28,32 @@ import (
 
 	"github.com/snapcore/snapd/gadget/quantity"
 	"github.com/snapcore/snapd/kernel"
+	"github.com/snapcore/snapd/secboot"
 )
 
-// LayoutConstraints defines the constraints for arranging structures within a
-// volume
-type LayoutConstraints struct {
-	// NonMBRStartOffset is the default start offset of non-MBR structure in
-	// the volume.
-	NonMBRStartOffset quantity.Offset
+// LayoutOptions defines the options to layout a given volume.
+type LayoutOptions struct {
 	// SkipResolveContent will skip resolving content paths
 	// and `$kernel:` style references
 	SkipResolveContent bool
+
+	// IgnoreContent will skip laying out content structure data to the
+	// volume. Settings this implies "SkipResolveContent".  This
+	// is used when only the partitions need to get
+	// created and content gets written later.
+	IgnoreContent bool
+
+	GadgetRootDir string
+	KernelRootDir string
+
+	EncType secboot.EncryptionType
 }
 
+// NonMBRStartOffset is the minimum start offset of the first non-MBR structure
+// in the volume that does not specify explicitly an offset. It can be ignored
+// by setting explicitly offsets.
+const NonMBRStartOffset = 1 * quantity.OffsetMiB
+
 // LaidOutVolume defines the size of a volume and arrangement of all the
 // structures within it
 type LaidOutVolume struct {
@@ -50,8 +63,6 @@ type LaidOutVolume struct {
 	// LaidOutStructure is a list of structures within the volume, sorted
 	// by their start offsets
 	LaidOutStructure []LaidOutStructure
-	// RootDir is the root directory for volume data
-	RootDir string
 }
 
 // PartiallyLaidOutVolume defines the layout of volume structures, but lacks the
@@ -63,18 +74,18 @@ type PartiallyLaidOutVolume struct {
 	LaidOutStructure []LaidOutStructure
 }
 
-// LaidOutStructure describes a VolumeStructure that has been placed within the
-// volume
+// LaidOutStructure describes a VolumeStructure coming from the gadget plus the
+// OnDiskStructure that describes how it would be applied to a given disk and
+// additional content used when writing/updating data in the structure.
 type LaidOutStructure struct {
-	*VolumeStructure
-	// StartOffset defines the start offset of the structure within the
-	// enclosing volume
-	StartOffset quantity.Offset
+	OnDiskStructure
+	// VolumeStructure is the volume structure defined in gadget.yaml
+	VolumeStructure *VolumeStructure
 	// AbsoluteOffsetWrite is the resolved absolute position of offset-write
 	// for this structure element within the enclosing volume
 	AbsoluteOffsetWrite *quantity.Offset
 	// Index of the structure definition in gadget YAML, note this starts at 0.
-	Index int
+	YamlIndex int
 	// LaidOutContent is a list of raw content inside the structure
 	LaidOutContent []LaidOutContent
 	// ResolvedContent is a list of filesystem content that has all
@@ -86,11 +97,56 @@ type LaidOutStructure struct {
 // meh this function is weirdly placed, not sure what to do w/o making schemaMBR
 // constant exported
 func IsRoleMBR(ls LaidOutStructure) bool {
-	return ls.Role == schemaMBR
+	return ls.Role() == schemaMBR
+}
+
+// These accessors return currently what comes in the gadget, but will use
+// OnDiskVolume data when the latter is made part of LaidOutStructure.
+
+// Type returns the type of the structure, which can be 2-hex digit MBR
+// partition, 36-char GUID partition, comma separated <mbr>,<guid> for hybrid
+// partitioning schemes, or 'bare' when the structure is not considered a
+// partition.
+//
+// For backwards compatibility type 'mbr' can also be returned, and
+// that is equivalent to role 'mbr'.
+func (l LaidOutStructure) Type() string {
+	return l.VolumeStructure.Type
+}
+
+// Name returns the partition label.
+func (l LaidOutStructure) Name() string {
+	return l.VolumeStructure.Name
+}
+
+// Label returns the filesystem label.
+func (l LaidOutStructure) Label() string {
+	return l.VolumeStructure.Label
+}
+
+// Filesystem for formatting the structure.
+func (l LaidOutStructure) Filesystem() string {
+	return l.VolumeStructure.Filesystem
+}
+
+// Role for the structure as specified in the gadget.
+func (l LaidOutStructure) Role() string {
+	return l.VolumeStructure.Role
+}
+
+// HasFilesystem returns true if the structure is using a filesystem.
+func (l *LaidOutStructure) HasFilesystem() bool {
+	return l.VolumeStructure.HasFilesystem()
+}
+
+// IsPartition returns true when the structure describes a partition in a block
+// device.
+func (l *LaidOutStructure) IsPartition() bool {
+	return l.VolumeStructure.IsPartition()
 }
 
 func (p LaidOutStructure) String() string {
-	return fmtIndexAndName(p.Index, p.Name)
+	return fmtIndexAndName(p.YamlIndex, p.Name())
 }
 
 type byStartOffset []LaidOutStructure
@@ -102,8 +158,8 @@ func (b byStartOffset) Less(i, j int) bool { return b[i].StartOffset < b[j].Star
 // LaidOutContent describes raw content that has been placed within the
 // encompassing structure and volume
 //
-// TODO: this can't have "$kernel:" refs at this point, fail in validate
-//       for bare structures with "$kernel:" refs
+// TODO: this can't have "$kernel:" refs at this point, fail in validate for
+// bare structures with "$kernel:" refs
 type LaidOutContent struct {
 	*VolumeContent
 
@@ -137,50 +193,43 @@ type ResolvedContent struct {
 	KernelUpdate bool
 }
 
-func layoutVolumeStructures(volume *Volume, constraints LayoutConstraints) (structures []LaidOutStructure, byName map[string]*LaidOutStructure, err error) {
-	previousEnd := quantity.Offset(0)
+func layoutVolumeStructures(volume *Volume) (structures []LaidOutStructure, byName map[string]*LaidOutStructure, err error) {
 	structures = make([]LaidOutStructure, len(volume.Structure))
 	byName = make(map[string]*LaidOutStructure, len(volume.Structure))
 
-	for idx, s := range volume.Structure {
-		var start quantity.Offset
-		if s.Offset == nil {
-			if s.Role != schemaMBR && previousEnd < constraints.NonMBRStartOffset {
-				start = constraints.NonMBRStartOffset
-			} else {
-				start = previousEnd
-			}
-		} else {
-			start = *s.Offset
-		}
-
-		end := start + quantity.Offset(s.Size)
+	for idx := range volume.Structure {
 		ps := LaidOutStructure{
 			VolumeStructure: &volume.Structure[idx],
-			StartOffset:     start,
-			Index:           idx,
+			YamlIndex:       idx,
 		}
 
-		if ps.Name != "" {
-			byName[ps.Name] = &ps
+		if ps.Name() != "" {
+			byName[ps.Name()] = &ps
+		}
+		// Fill the parts of OnDiskStructure that do not depend on the disk
+		// or on whether we are encrypting or not.
+		// TODO Eventually fill everything here by passing all needed info
+		ps.OnDiskStructure = OnDiskStructure{
+			Name:        ps.VolumeStructure.Name,
+			Type:        ps.VolumeStructure.Type,
+			StartOffset: *volume.Structure[idx].Offset,
+			Size:        ps.VolumeStructure.Size,
 		}
 
 		structures[idx] = ps
-
-		previousEnd = end
 	}
 
 	// sort by starting offset
 	sort.Sort(byStartOffset(structures))
 
-	previousEnd = quantity.Offset(0)
+	previousEnd := quantity.Offset(0)
 	for idx, ps := range structures {
 		if ps.StartOffset < previousEnd {
 			return nil, nil, fmt.Errorf("cannot lay out volume, structure %v overlaps with preceding structure %v", ps, structures[idx-1])
 		}
-		previousEnd = ps.StartOffset + quantity.Offset(ps.Size)
+		previousEnd = ps.StartOffset + quantity.Offset(ps.VolumeStructure.Size)
 
-		offsetWrite, err := resolveOffsetWrite(ps.OffsetWrite, byName)
+		offsetWrite, err := resolveOffsetWrite(ps.VolumeStructure.OffsetWrite, byName)
 		if err != nil {
 			return nil, nil, fmt.Errorf("cannot resolve offset-write of structure %v: %v", ps, err)
 		}
@@ -190,9 +239,9 @@ func layoutVolumeStructures(volume *Volume, constraints LayoutConstraints) (stru
 	return structures, byName, nil
 }
 
-// LayoutVolumePartially attempts to lay out only the structures in the volume using provided constraints
-func LayoutVolumePartially(volume *Volume, constraints LayoutConstraints) (*PartiallyLaidOutVolume, error) {
-	structures, _, err := layoutVolumeStructures(volume, constraints)
+// LayoutVolumePartially attempts to lay out only the structures in the volume.
+func LayoutVolumePartially(volume *Volume) (*PartiallyLaidOutVolume, error) {
+	structures, _, err := layoutVolumeStructures(volume)
 	if err != nil {
 		return nil, err
 	}
@@ -204,13 +253,31 @@ func LayoutVolumePartially(volume *Volume, constraints LayoutConstraints) (*Part
 	return vol, nil
 }
 
+func setOnDiskLabelAndTypeInLaidOuts(los []LaidOutStructure, encType secboot.EncryptionType) {
+	for i := range los {
+		los[i].PartitionFSLabel = los[i].Label()
+		los[i].PartitionFSType = los[i].Filesystem()
+		if encType != secboot.EncryptionTypeNone {
+			switch los[i].Role() {
+			case SystemData, SystemSave:
+				los[i].PartitionFSLabel += "-enc"
+				los[i].PartitionFSType = "crypto_LUKS"
+			}
+		}
+	}
+}
+
 // LayoutVolume attempts to completely lay out the volume, that is the
-// structures and their content, using provided constraints
-func LayoutVolume(gadgetRootDir, kernelRootDir string, volume *Volume, constraints LayoutConstraints) (*LaidOutVolume, error) {
+// structures and their content, using provided options.
+func LayoutVolume(volume *Volume, opts *LayoutOptions) (*LaidOutVolume, error) {
 	var err error
+	if opts == nil {
+		opts = &LayoutOptions{}
+	}
+	doResolveContent := !(opts.IgnoreContent || opts.SkipResolveContent)
 
 	var kernelInfo *kernel.Info
-	if !constraints.SkipResolveContent {
+	if doResolveContent {
 		// TODO:UC20: check and error if kernelRootDir == "" here
 		// This needs the upper layer of gadget updates to be
 		// updated to pass the kernel root first.
@@ -218,13 +285,13 @@ func LayoutVolume(gadgetRootDir, kernelRootDir string, volume *Volume, constrain
 		// Note that the kernelRootDir may reference the running
 		// kernel if there is a gadget update or the new kernel if
 		// there is a kernel update.
-		kernelInfo, err = kernel.ReadInfo(kernelRootDir)
+		kernelInfo, err = kernel.ReadInfo(opts.KernelRootDir)
 		if err != nil {
 			return nil, err
 		}
 	}
 
-	structures, byName, err := layoutVolumeStructures(volume, constraints)
+	structures, byName, err := layoutVolumeStructures(volume)
 	if err != nil {
 		return nil, err
 	}
@@ -236,27 +303,34 @@ func LayoutVolume(gadgetRootDir, kernelRootDir string, volume *Volume, constrain
 		if ps.AbsoluteOffsetWrite != nil && *ps.AbsoluteOffsetWrite > fartherstOffsetWrite {
 			fartherstOffsetWrite = *ps.AbsoluteOffsetWrite
 		}
-		if end := ps.StartOffset + quantity.Offset(ps.Size); end > farthestEnd {
+		if end := ps.StartOffset + quantity.Offset(ps.VolumeStructure.Size); end > farthestEnd {
 			farthestEnd = end
 		}
 
-		// lay out raw content
-		content, err := layOutStructureContent(gadgetRootDir, &structures[idx], byName)
-		if err != nil {
-			return nil, err
-		}
+		// Set appropriately label and type details
+		// TODO: set this in layoutVolumeStructures in the future.
+		setOnDiskLabelAndTypeInLaidOuts(structures, opts.EncType)
 
-		for _, c := range content {
-			if c.AbsoluteOffsetWrite != nil && *c.AbsoluteOffsetWrite > fartherstOffsetWrite {
-				fartherstOffsetWrite = *c.AbsoluteOffsetWrite
+		// Lay out raw content. This can be skipped when only partition
+		// creation is needed and is safe because each volume structure
+		// has a size so even without the structure content the layout
+		// can be calculated.
+		if !opts.IgnoreContent {
+			content, err := layOutStructureContent(opts.GadgetRootDir, &structures[idx], byName)
+			if err != nil {
+				return nil, err
 			}
+			for _, c := range content {
+				if c.AbsoluteOffsetWrite != nil && *c.AbsoluteOffsetWrite > fartherstOffsetWrite {
+					fartherstOffsetWrite = *c.AbsoluteOffsetWrite
+				}
+			}
+			structures[idx].LaidOutContent = content
 		}
 
-		structures[idx].LaidOutContent = content
-
 		// resolve filesystem content
-		if !constraints.SkipResolveContent {
-			resolvedContent, err := resolveVolumeContent(gadgetRootDir, kernelRootDir, kernelInfo, &structures[idx], nil)
+		if doResolveContent {
+			resolvedContent, err := resolveVolumeContent(opts.GadgetRootDir, opts.KernelRootDir, kernelInfo, &structures[idx], nil)
 			if err != nil {
 				return nil, err
 			}
@@ -273,7 +347,6 @@ func LayoutVolume(gadgetRootDir, kernelRootDir string, volume *Volume, constrain
 		Volume:           volume,
 		Size:             volumeSize,
 		LaidOutStructure: structures,
-		RootDir:          gadgetRootDir,
 	}
 	return vol, nil
 }
@@ -283,18 +356,18 @@ func resolveVolumeContent(gadgetRootDir, kernelRootDir string, kernelInfo *kerne
 		// structures without a file system are not resolved here
 		return nil, nil
 	}
-	if len(ps.Content) == 0 {
+	if len(ps.VolumeStructure.Content) == 0 {
 		return nil, nil
 	}
 
-	content := make([]ResolvedContent, 0, len(ps.Content))
-	for idx := range ps.Content {
-		resolvedSource, kupdate, err := resolveContentPathOrRef(gadgetRootDir, kernelRootDir, kernelInfo, ps.Content[idx].UnresolvedSource)
+	content := make([]ResolvedContent, 0, len(ps.VolumeStructure.Content))
+	for idx := range ps.VolumeStructure.Content {
+		resolvedSource, kupdate, err := resolveContentPathOrRef(gadgetRootDir, kernelRootDir, kernelInfo, ps.VolumeStructure.Content[idx].UnresolvedSource)
 		if err != nil {
 			return nil, fmt.Errorf("cannot resolve content for structure %v at index %v: %v", ps, idx, err)
 		}
 		rc := ResolvedContent{
-			VolumeContent:  &ps.Content[idx],
+			VolumeContent:  &ps.VolumeStructure.Content[idx],
 			ResolvedSource: resolvedSource,
 			KernelUpdate:   kupdate,
 		}
@@ -388,14 +461,14 @@ func layOutStructureContent(gadgetRootDir string, ps *LaidOutStructure, known ma
 		// structures with a filesystem do not need any extra layout
 		return nil, nil
 	}
-	if len(ps.Content) == 0 {
+	if len(ps.VolumeStructure.Content) == 0 {
 		return nil, nil
 	}
 
-	content := make([]LaidOutContent, len(ps.Content))
+	content := make([]LaidOutContent, len(ps.VolumeStructure.Content))
 	previousEnd := quantity.Offset(0)
 
-	for idx, c := range ps.Content {
+	for idx, c := range ps.VolumeStructure.Content {
 		imageSize, err := getImageSize(filepath.Join(gadgetRootDir, c.Image))
 		if err != nil {
 			return nil, fmt.Errorf("cannot lay out structure %v: content %q: %v", ps, c.Image, err)
@@ -423,7 +496,7 @@ func layOutStructureContent(gadgetRootDir string, ps *LaidOutStructure, known ma
 		}
 
 		content[idx] = LaidOutContent{
-			VolumeContent: &ps.Content[idx],
+			VolumeContent: &ps.VolumeStructure.Content[idx],
 			Size:          actualSize,
 			StartOffset:   ps.StartOffset + start,
 			Index:         idx,
@@ -431,7 +504,7 @@ func layOutStructureContent(gadgetRootDir string, ps *LaidOutStructure, known ma
 			AbsoluteOffsetWrite: offsetWrite,
 		}
 		previousEnd = start + quantity.Offset(actualSize)
-		if quantity.Size(previousEnd) > ps.Size {
+		if quantity.Size(previousEnd) > ps.VolumeStructure.Size {
 			return nil, fmt.Errorf("cannot lay out structure %v: content %q does not fit in the structure", ps, c.Image)
 		}
 	}
