@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2016-2017 Canonical Ltd
+ * Copyright (C) 2016-2022 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -23,8 +23,10 @@ import (
 	"encoding/json"
 	"errors"
 	"fmt"
+	"io"
 	"io/ioutil"
 	"os"
+	"os/exec"
 	"path/filepath"
 	"runtime"
 	"syscall"
@@ -34,10 +36,10 @@ import (
 	. "gopkg.in/check.v1"
 	"gopkg.in/tomb.v2"
 
+	"github.com/snapcore/snapd/boot"
 	"github.com/snapcore/snapd/dirs"
 	"github.com/snapcore/snapd/osutil"
 	"github.com/snapcore/snapd/overlord"
-	"github.com/snapcore/snapd/overlord/auth"
 	"github.com/snapcore/snapd/overlord/devicestate/devicestatetest"
 	"github.com/snapcore/snapd/overlord/hookstate"
 	"github.com/snapcore/snapd/overlord/ifacestate"
@@ -113,6 +115,7 @@ func (ovs *overlordSuite) TestNew(c *C) {
 
 	c.Check(o.StateEngine(), NotNil)
 	c.Check(o.TaskRunner(), NotNil)
+	c.Check(o.RestartManager(), NotNil)
 	c.Check(o.SnapManager(), NotNil)
 	c.Check(o.ServiceManager(), NotNil)
 	c.Check(o.AssertManager(), NotNil)
@@ -172,6 +175,7 @@ func (ovs *overlordSuite) TestNewWithGoodState(c *C) {
 
 	o, err := overlord.New(nil)
 	c.Assert(err, IsNil)
+	c.Check(o.RestartManager(), NotNil)
 
 	state := o.State()
 	c.Assert(err, IsNil)
@@ -306,12 +310,7 @@ func (wm *witnessManager) Ensure() error {
 func markSeeded(o *overlord.Overlord) {
 	st := o.State()
 	st.Lock()
-	st.Set("seeded", true)
-	devicestatetest.SetDevice(st, &auth.DeviceState{
-		Brand:  "canonical",
-		Model:  "pc",
-		Serial: "serialserial",
-	})
+	devicestatetest.MarkInitialized(st)
 	st.Unlock()
 }
 
@@ -670,9 +669,9 @@ func (ovs *overlordSuite) TestOverlordStartUpSetsStartOfOperation(c *C) {
 	st.Lock()
 	defer st.Unlock()
 
-	// sanity check, not set
+	// validity check, not set
 	var opTime time.Time
-	c.Assert(st.Get("start-of-operation-time", &opTime), Equals, state.ErrNoState)
+	c.Assert(st.Get("start-of-operation-time", &opTime), testutil.ErrorIs, state.ErrNoState)
 	st.Unlock()
 
 	c.Assert(o.StartUp(), IsNil)
@@ -713,7 +712,7 @@ func (ovs *overlordSuite) TestEnsureLoopPruneDoesntAbortShortlyAfterStartOfOpera
 
 	restoreTimeNow()
 
-	// sanity
+	// validity
 	c.Check(st.Changes(), HasLen, 1)
 
 	st.Unlock()
@@ -770,7 +769,7 @@ func (ovs *overlordSuite) TestEnsureLoopPruneAbortsOld(c *C) {
 
 	restoreTimeNow()
 
-	// sanity
+	// validity
 	c.Check(st.Changes(), HasLen, 1)
 	st.Unlock()
 
@@ -783,7 +782,7 @@ func (ovs *overlordSuite) TestEnsureLoopPruneAbortsOld(c *C) {
 	st.Lock()
 	defer st.Unlock()
 
-	// sanity
+	// validity
 	op, err := o.DeviceManager().StartOfOperationTime()
 	c.Assert(err, IsNil)
 	c.Check(op.Equal(opTime), Equals, true)
@@ -812,7 +811,7 @@ func (ovs *overlordSuite) TestEnsureLoopNoPruneWhenPreseed(c *C) {
 		return &state.Retry{}
 	}, nil)
 
-	// sanity
+	// validity
 	_, err = o.DeviceManager().StartOfOperationTime()
 	c.Assert(err, ErrorMatches, `internal error: unexpected call to StartOfOperationTime in preseed mode`)
 
@@ -840,7 +839,7 @@ func (ovs *overlordSuite) TestEnsureLoopNoPruneWhenPreseed(c *C) {
 	defer st.Unlock()
 
 	var opTime time.Time
-	c.Assert(st.Get("start-of-operation-time", &opTime), Equals, state.ErrNoState)
+	c.Assert(st.Get("start-of-operation-time", &opTime), testutil.ErrorIs, state.ErrNoState)
 	c.Check(chg.Status(), Equals, state.DoingStatus)
 }
 
@@ -1138,7 +1137,7 @@ func (ovs *overlordSuite) TestRequestRestartNoHandler(c *C) {
 	st.Lock()
 	defer st.Unlock()
 
-	restart.Request(st, restart.RestartDaemon)
+	restart.Request(st, restart.RestartDaemon, nil)
 }
 
 type testRestartHandler struct {
@@ -1147,7 +1146,7 @@ type testRestartHandler struct {
 	rebootVerifiedErr error
 }
 
-func (rb *testRestartHandler) HandleRestart(t restart.RestartType) {
+func (rb *testRestartHandler) HandleRestart(t restart.RestartType, ri *boot.RebootInfo) {
 	rb.restartRequested = t
 }
 
@@ -1171,7 +1170,7 @@ func (ovs *overlordSuite) TestRequestRestartHandler(c *C) {
 	st.Lock()
 	defer st.Unlock()
 
-	restart.Request(st, restart.RestartDaemon)
+	restart.Request(st, restart.RestartDaemon, nil)
 
 	c.Check(rb.restartRequested, Equals, restart.RestartDaemon)
 }
@@ -1315,3 +1314,65 @@ func (ovs *overlordSuite) TestStartupTimeout(c *C) {
 	c.Check(to, Equals, (30+5+5)*time.Second)
 	c.Check(reasoning, Equals, "pessimistic estimate of 30s plus 5s per snap")
 }
+
+func (ovs *overlordSuite) TestLockWithTimeoutHappy(c *C) {
+	f, err := ioutil.TempFile("", "testlock-*")
+	defer func() {
+		f.Close()
+		os.Remove(f.Name())
+	}()
+	c.Assert(err, IsNil)
+	flock, err := osutil.NewFileLock(f.Name())
+	c.Assert(err, IsNil)
+
+	err = overlord.LockWithTimeout(flock, time.Second)
+	c.Check(err, IsNil)
+}
+
+func (ovs *overlordSuite) TestLockWithTimeoutFailed(c *C) {
+	// Set the state lock retry interval to 0.1 ms; the timeout is not used in
+	// this test (we specify the timeout when calling the lockWithTimeout()
+	// function below); we set it to a big value in order to trigger a test
+	// failure in case the logic gets modified and it suddenly becomes
+	// relevant.
+	restoreTimeout := overlord.MockStateLockTimeout(time.Hour, 100*time.Microsecond)
+	defer restoreTimeout()
+
+	var notifyCalls []string
+	restoreNotify := overlord.MockSystemdSdNotify(func(notifyState string) error {
+		notifyCalls = append(notifyCalls, notifyState)
+		return nil
+	})
+	defer restoreNotify()
+
+	f, err := ioutil.TempFile("", "testlock-*")
+	defer func() {
+		f.Close()
+		os.Remove(f.Name())
+	}()
+	c.Assert(err, IsNil)
+	flock, err := osutil.NewFileLock(f.Name())
+	c.Assert(err, IsNil)
+
+	cmd := exec.Command("flock", "-w", "2", f.Name(), "-c", "echo acquired && sleep 5")
+	stdout, err := cmd.StdoutPipe()
+	c.Assert(err, IsNil)
+	err = cmd.Start()
+	c.Assert(err, IsNil)
+	defer func() {
+		cmd.Process.Kill()
+		cmd.Wait()
+	}()
+
+	// Wait until the shell command prints "acquired"
+	buf := make([]byte, 8)
+	bytesRead, err := io.ReadAtLeast(stdout, buf, len(buf))
+	c.Assert(err, IsNil)
+	c.Assert(bytesRead, Equals, len(buf))
+
+	err = overlord.LockWithTimeout(flock, 5*time.Millisecond)
+	c.Check(err, ErrorMatches, "timeout for state lock file expired")
+	c.Check(notifyCalls, DeepEquals, []string{
+		"EXTEND_TIMEOUT_USEC=5000",
+	})
+}
