@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2014-2021 Canonical Ltd
+ * Copyright (C) 2014-2022 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -20,6 +20,7 @@
 package snap_test
 
 import (
+	"errors"
 	"fmt"
 	"io/ioutil"
 	"os"
@@ -106,7 +107,7 @@ func (s *infoSuite) TestContactFromEdited(c *C) {
 	}
 
 	info.SideInfo = snap.SideInfo{
-		EditedContact: "mailto:econtact",
+		LegacyEditedContact: "mailto:econtact",
 	}
 
 	c.Check(info.Contact(), Equals, "mailto:econtact")
@@ -152,29 +153,68 @@ func (s *infoSuite) TestLinks(c *C) {
 	info := &snap.Info{
 		OriginalLinks: map[string][]string{
 			"contact": {"ocontact"},
-			"website": {"owebsite"},
+			"website": {"http://owebsite"},
 		},
 	}
 
 	info.SideInfo = snap.SideInfo{
 		EditedLinks: map[string][]string{
-			"contact": {"econtact"},
-			"website": {"ewebsite"},
+			"contact": {"mailto:econtact"},
+			"website": {"http://ewebsite"},
 		},
 	}
 
 	c.Check(info.Links(), DeepEquals, map[string][]string{
-		"contact": {"econtact"},
-		"website": {"ewebsite"},
+		"contact": {"mailto:econtact"},
+		"website": {"http://ewebsite"},
 	})
 
 	info.EditedLinks = nil
 	c.Check(info.Links(), DeepEquals, map[string][]string{
-		"contact": {"ocontact"},
-		"website": {"owebsite"},
+		"contact": {"mailto:ocontact"},
+		"website": {"http://owebsite"},
 	})
 }
 
+func (s *infoSuite) TestNormalizeOriginalLinks(c *C) {
+	info := &snap.Info{
+		OriginalLinks: map[string][]string{
+			"contact": {"ocontact", "mailto:ocontact"},
+			"website": {":", "http://owebsite", ""},
+		},
+	}
+
+	c.Check(info.Links(), DeepEquals, map[string][]string{
+		"contact": {"mailto:ocontact"},
+		"website": {"http://owebsite"},
+	})
+}
+
+func (s *infoSuite) TestWebsiteFromLegacy(c *C) {
+	info := &snap.Info{
+		OriginalLinks: nil,
+		LegacyWebsite: "http://website",
+	}
+
+	c.Check(info.Website(), Equals, "http://website")
+}
+
+func (s *infoSuite) TestNoWebsite(c *C) {
+	info := &snap.Info{}
+
+	c.Check(info.Website(), Equals, "")
+}
+
+func (s *infoSuite) TestWebsiteFromLinks(c *C) {
+	info := &snap.Info{
+		OriginalLinks: map[string][]string{
+			"website": {"http://website1", "http://website2"},
+		},
+	}
+
+	c.Check(info.Website(), Equals, "http://website1")
+}
+
 func (s *infoSuite) TestAppInfoSecurityTag(c *C) {
 	appInfo := &snap.AppInfo{Snap: &snap.Info{SuggestedName: "http"}, Name: "GET"}
 	c.Check(appInfo.SecurityTag(), Equals, "snap.http.GET")
@@ -349,7 +389,7 @@ func (s *infoSuite) TestInstallDate(c *C) {
 	si := &snap.SideInfo{Revision: snap.R(1)}
 	info := snaptest.MockSnap(c, sampleYaml, si)
 	// not current -> Zero
-	c.Check(info.InstallDate().IsZero(), Equals, true)
+	c.Check(info.InstallDate(), IsNil)
 	c.Check(snap.InstallDate(info.InstanceName()).IsZero(), Equals, true)
 
 	mountdir := info.MountDir()
@@ -360,7 +400,7 @@ func (s *infoSuite) TestInstallDate(c *C) {
 	st, err := os.Lstat(cur)
 	c.Assert(err, IsNil)
 	instTime := st.ModTime()
-	// sanity
+	// validity
 	c.Check(instTime.IsZero(), Equals, false)
 
 	c.Check(info.InstallDate().Equal(instTime), Equals, true)
@@ -586,6 +626,23 @@ confinement: foo`
 	c.Assert(err, ErrorMatches, ".*invalid confinement type.*")
 }
 
+func (s *infoSuite) TestReadInfoFromSnapFileChatchesInvalidSnapshot(c *C) {
+	yaml := `name: foo
+version: 1.0
+type: app`
+	contents := [][]string{
+		{"meta/snapshots.yaml", "Oops! This is not really valid yaml :-("},
+	}
+	sideInfo := &snap.SideInfo{}
+	snapInfo := snaptest.MockSnapWithFiles(c, yaml, sideInfo, contents)
+
+	snapf, err := snapfile.Open(snapInfo.MountDir())
+	c.Assert(err, IsNil)
+
+	_, err = snap.ReadInfoFromSnapFile(snapf, nil)
+	c.Assert(err, ErrorMatches, "cannot read snapshot manifest: yaml: unmarshal errors:\n.*")
+}
+
 func (s *infoSuite) TestAppEnvSimple(c *C) {
 	yaml := `name: foo
 version: 1.0
@@ -1087,6 +1144,7 @@ func (s *infoSuite) testDirAndFileMethods(c *C, info snap.PlaceInfo) {
 	c.Check(info.UserDataDir("/home/bob", nil), Equals, "/home/bob/snap/name/1")
 	c.Check(info.UserCommonDataDir("/home/bob", nil), Equals, "/home/bob/snap/name/common")
 	c.Check(info.CommonDataDir(), Equals, "/var/snap/name/common")
+	c.Check(info.CommonDataSaveDir(), Equals, "/var/lib/snapd/save/snap/name")
 	c.Check(info.UserXdgRuntimeDir(12345), Equals, "/run/user/12345/snap.name")
 	// XXX: Those are actually a globs, not directories
 	c.Check(info.DataHomeDir(nil), Equals, "/home/*/snap/name/1")
@@ -1115,6 +1173,7 @@ func (s *infoSuite) testInstanceDirAndFileMethods(c *C, info snap.PlaceInfo) {
 	c.Check(info.UserDataDir("/home/bob", nil), Equals, "/home/bob/snap/name_instance/1")
 	c.Check(info.UserCommonDataDir("/home/bob", nil), Equals, "/home/bob/snap/name_instance/common")
 	c.Check(info.CommonDataDir(), Equals, "/var/snap/name_instance/common")
+	c.Check(info.CommonDataSaveDir(), Equals, "/var/lib/snapd/save/snap/name_instance")
 	c.Check(info.UserXdgRuntimeDir(12345), Equals, "/run/user/12345/snap.name_instance")
 	// XXX: Those are actually a globs, not directories
 	c.Check(info.DataHomeDir(nil), Equals, "/home/*/snap/name_instance/1")
@@ -1412,7 +1471,7 @@ func (s *infoSuite) TestPlugInfoAttr(c *C) {
 	c.Assert(plug.Attr("number", &intVal), IsNil)
 	c.Check(intVal, Equals, 123)
 
-	c.Check(plug.Attr("key", &intVal), ErrorMatches, `snap "snap" has interface "interface" with invalid value type for "key" attribute`)
+	c.Check(plug.Attr("key", &intVal), ErrorMatches, `snap "snap" has interface "interface" with invalid value type string for "key" attribute: \*int`)
 	c.Check(plug.Attr("unknown", &val), ErrorMatches, `snap "snap" does not have attribute "unknown" for interface "interface"`)
 	c.Check(plug.Attr("key", intVal), ErrorMatches, `internal error: cannot get "key" attribute of interface "interface" with non-pointer value`)
 }
@@ -1429,7 +1488,7 @@ func (s *infoSuite) TestSlotInfoAttr(c *C) {
 	c.Assert(slot.Attr("number", &intVal), IsNil)
 	c.Check(intVal, Equals, 123)
 
-	c.Check(slot.Attr("key", &intVal), ErrorMatches, `snap "snap" has interface "interface" with invalid value type for "key" attribute`)
+	c.Check(slot.Attr("key", &intVal), ErrorMatches, `snap "snap" has interface "interface" with invalid value type string for "key" attribute: \*int`)
 	c.Check(slot.Attr("unknown", &val), ErrorMatches, `snap "snap" does not have attribute "unknown" for interface "interface"`)
 	c.Check(slot.Attr("key", intVal), ErrorMatches, `internal error: cannot get "key" attribute of interface "interface" with non-pointer value`)
 }
@@ -1546,6 +1605,8 @@ func (s *infoSuite) TestStopModeTypeKillMode(c *C) {
 		{"sigusr1-all", true},
 		{"sigusr2", false},
 		{"sigusr2-all", true},
+		{"sigint", false},
+		{"sigint-all", true},
 	} {
 		c.Check(snap.StopModeType(t.stopMode).KillAll(), Equals, t.killall, Commentf("wrong KillAll for %v", t.stopMode))
 	}
@@ -1652,6 +1713,7 @@ func (s *infoSuite) TestDirAndFileHelpers(c *C) {
 	c.Check(snap.HooksDir("name", snap.R(1)), Equals, fmt.Sprintf("%s/name/1/meta/hooks", dirs.SnapMountDir))
 	c.Check(snap.DataDir("name", snap.R(1)), Equals, "/var/snap/name/1")
 	c.Check(snap.CommonDataDir("name"), Equals, "/var/snap/name/common")
+	c.Check(snap.CommonDataSaveDir("name"), Equals, "/var/lib/snapd/save/snap/name")
 	c.Check(snap.UserDataDir("/home/bob", "name", snap.R(1), nil), Equals, "/home/bob/snap/name/1")
 	c.Check(snap.UserCommonDataDir("/home/bob", "name", nil), Equals, "/home/bob/snap/name/common")
 	c.Check(snap.UserXdgRuntimeDir(12345, "name"), Equals, "/run/user/12345/snap.name")
@@ -1662,6 +1724,7 @@ func (s *infoSuite) TestDirAndFileHelpers(c *C) {
 	c.Check(snap.HooksDir("name_instance", snap.R(1)), Equals, fmt.Sprintf("%s/name_instance/1/meta/hooks", dirs.SnapMountDir))
 	c.Check(snap.DataDir("name_instance", snap.R(1)), Equals, "/var/snap/name_instance/1")
 	c.Check(snap.CommonDataDir("name_instance"), Equals, "/var/snap/name_instance/common")
+	c.Check(snap.CommonDataSaveDir("name_instance"), Equals, "/var/lib/snapd/save/snap/name_instance")
 	c.Check(snap.UserDataDir("/home/bob", "name_instance", snap.R(1), nil), Equals, "/home/bob/snap/name_instance/1")
 	c.Check(snap.UserCommonDataDir("/home/bob", "name_instance", nil), Equals, "/home/bob/snap/name_instance/common")
 	c.Check(snap.UserXdgRuntimeDir(12345, "name_instance"), Equals, "/run/user/12345/snap.name_instance")
@@ -1863,3 +1926,125 @@ func (s *infoSuite) TestHelpersWithHiddenSnapFolder(c *C) {
 	c.Check(snap.UserCommonDataDir("/home/bob", "name_instance", opts), Equals, "/home/bob/.snap/data/name_instance/common")
 	c.Check(snap.UserSnapDir("/home/bob", "name_instance", opts), Equals, "/home/bob/.snap/data/name_instance")
 }
+
+func (s *infoSuite) TestGetAttributeUnhappy(c *C) {
+	attrs := map[string]interface{}{}
+	var stringVal string
+	err := snap.GetAttribute("snap0", "iface0", attrs, "non-existent", &stringVal)
+	c.Check(stringVal, Equals, "")
+	c.Check(err, ErrorMatches, `snap "snap0" does not have attribute "non-existent" for interface "iface0"`)
+	c.Check(errors.Is(err, snap.AttributeNotFoundError{}), Equals, true)
+}
+
+func (s *infoSuite) TestGetAttributeHappy(c *C) {
+	attrs := map[string]interface{}{
+		"attr0": "a string",
+		"attr1": 12,
+	}
+	var intVal int
+	err := snap.GetAttribute("snap0", "iface0", attrs, "attr1", &intVal)
+	c.Check(err, IsNil)
+	c.Check(intVal, Equals, 12)
+}
+
+func (s *infoSuite) TestSnapdAssertionMaxFormatsFromSnapFileFromSnapd(c *C) {
+	tests := []struct {
+		info     string
+		snapDecl int
+		sysUser  int
+	}{
+		{info: `VERSION=2.58
+SNAPD_ASSERTS_FORMATS='{"snap-declaration":5,"system-user":2}'`, snapDecl: 5, sysUser: 2},
+		{info: `VERSION=2.56
+SNAPD_ASSERTS_FORMATS='{"snap-declaration":5,"system-user":1}'`, snapDecl: 5, sysUser: 1},
+		{info: `VERSION=2.55`, snapDecl: 5, sysUser: 1},
+		{info: `VERSION=2.54`, snapDecl: 5, sysUser: 1},
+		{info: `VERSION=2.47`, snapDecl: 4, sysUser: 1},
+		{info: `VERSION=2.46`, snapDecl: 4, sysUser: 1},
+		{info: `VERSION=2.45`, snapDecl: 4},
+		{info: `VERSION=2.44`, snapDecl: 4},
+		{info: `VERSION=2.36`, snapDecl: 3},
+		// old
+		{info: `VERSION=2.23`, snapDecl: 2},
+		// ancient
+		{info: `VERSION=2.17`, snapDecl: 1},
+		{info: `VERSION=2.16`},
+	}
+	for _, t := range tests {
+		snapdPath := snaptest.MakeTestSnapWithFiles(c, `name: snapd
+type: snapd
+version: 1.0`, [][]string{{
+			"/usr/lib/snapd/info", t.info}})
+		snapf, err := snapfile.Open(snapdPath)
+		c.Assert(err, IsNil)
+
+		maxFormats, ver, err := snap.SnapdAssertionMaxFormatsFromSnapFile(snapf)
+		c.Assert(err, IsNil)
+		expectedMaxFormats := map[string]int{}
+		if t.sysUser > 0 {
+			expectedMaxFormats["system-user"] = t.sysUser
+		}
+		if t.snapDecl > 0 {
+			expectedMaxFormats["snap-declaration"] = t.snapDecl
+		}
+		c.Check(maxFormats, DeepEquals, expectedMaxFormats)
+		c.Check(strings.HasPrefix(t.info, fmt.Sprintf("VERSION=%s", ver)), Equals, true)
+	}
+}
+
+func (s *infoSuite) TestSnapdAssertionMaxFormatsFromSnapFileFromCore(c *C) {
+	corePath := snaptest.MakeTestSnapWithFiles(c, `name: core
+type: os
+version: 1.0`, [][]string{{
+		"/usr/lib/snapd/info", `VERSION=2.47`}})
+	snapf, err := snapfile.Open(corePath)
+	c.Assert(err, IsNil)
+
+	maxFormats, ver, err := snap.SnapdAssertionMaxFormatsFromSnapFile(snapf)
+	c.Assert(err, IsNil)
+	c.Check(ver, Equals, "2.47")
+	c.Check(maxFormats, DeepEquals, map[string]int{
+		"snap-declaration": 4,
+		"system-user":      1,
+	})
+}
+
+func (s *infoSuite) TestSnapdAssertionMaxFormatsFromSnapFileFromKernel(c *C) {
+	krnlPath := snaptest.MakeTestSnapWithFiles(c, `name: krnl
+type: kernel
+version: 1.0`, [][]string{{
+		"/snapd-info", `VERSION=2.56
+SNAPD_ASSERTS_FORMATS='{"snap-declaration":5,"system-user":1}'`}})
+	snapf, err := snapfile.Open(krnlPath)
+	c.Assert(err, IsNil)
+
+	maxFormats, ver, err := snap.SnapdAssertionMaxFormatsFromSnapFile(snapf)
+	c.Assert(err, IsNil)
+	c.Check(ver, Equals, "2.56")
+	c.Check(maxFormats, DeepEquals, map[string]int{
+		"snap-declaration": 5,
+		"system-user":      1,
+	})
+
+	// no snadd-info
+	krnlPath = snaptest.MakeTestSnapWithFiles(c, `name: krnl
+type: kernel
+version: 1.0`, nil)
+	snapf, err = snapfile.Open(krnlPath)
+	c.Assert(err, IsNil)
+
+	maxFormats, ver, err = snap.SnapdAssertionMaxFormatsFromSnapFile(snapf)
+	c.Assert(err, IsNil)
+	c.Check(ver, Equals, "")
+	c.Check(maxFormats, IsNil)
+}
+
+func (s *infoSuite) TestSnapdAssertionMaxFormatsFromSnapFileFromOther(c *C) {
+	appPath := snaptest.MakeTestSnapWithFiles(c, `name: app
+version: 1.0`, nil)
+	snapf, err := snapfile.Open(appPath)
+	c.Assert(err, IsNil)
+
+	_, _, err = snap.SnapdAssertionMaxFormatsFromSnapFile(snapf)
+	c.Check(err, ErrorMatches, `cannot extract assertion max formats information, snaps of type app do not carry snapd`)
+}
